#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
__init__.py is a special Python file that allows a directory to become
a Python package so it can be accessed using the 'import' statement.
"""

from .cwb import Corpora, Corpus
from .version import __version__
